import matplotlib
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt

#todo fix the assertions in these
def plot_data(x, y, ax=False, line='r-', linethickness=0.5):
        if ax:
            #assert isinstance(ax, axes._subplots.AxesSubplot), "the figure passed isn't the correct format, please pass" \
                                                                 " an axes object"
        else:
            fig, ax = plt.subplots()

        ax.plot(x, y, line, linewidth=linethickness, alpha=0.5)
        return ax, plt

def plot_fits(x, peaks, ax=False, linethickness=0.5):
        if ax:
            #assert isinstance(ax, axes._subplots.AxesSubplot), "the figure passed isn't the correct format, please pass" \
                                                                " an axes object"
        else:
            fig, ax = plt.subplots()

        for i, peak in enumerate(peaks):
            ax.plot(x, peaks[peak], linewidth=linethickness)
        return ax, plt

def plot_background(x, background_data, ax=False, line='b--', linethickness=0.5):
        if ax:
            #assert isinstance(ax, axes._subplots.AxesSubplot), "the figure passed isn't the correct format, please pass" \
                                                                 " an axes object"
        else:
            fig, ax = plt.subplots()

        ax.plot(x, background_data, line, linewidth=linethickness)
        return ax, plt

def plot_fit_sum(x, peak_sum, background, ax=False, line='k-', linethickness=0.5): # option of including background
        if ax:
            #assert isinstance(ax, axes._subplots.AxesSubplot), "the figure passed isn't the correct format, please pass" \
                                                                 " an axes object"
        else:
            fig, ax = plt.subplots()

        sum = peak_sum + background

        ax.plot(x, sum, line, linewidth=linethickness)
        return ax, plt

def plot_uncertainty_curve(x, eval_unc, peak_sum, ax=False, color="#ABABAB"):
        if ax:
            #assert isinstance(ax, axes._subplots.AxesSubplot), "the figure passed isn't the correct format, please pass" \
                                                                 " an axes object"
        else:
            fig, ax = plt.subplots()

        ax.fill_between(x, peak_sum - eval_unc, peak_sum + eval_unc, color=color) #plot a grey band of uncertainty

        return ax, plt


from numpy import round
from glob import glob
from copy import deepcopy
from pandas.errors import ParserError
from tqdm import tqdm
from ast import literal_eval
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('TkAgg')
from typing import Union

from .thundobj import Thunder
from . import utilities as utili
from .background import background_removal as bg_remove
from . import peak_finding

############## NOT IMPLEMENTED!
# TODO
# make option of passing in many params files - one for each data file

class ThunderBag():

    def __init__(self, input):
        # initialise everything first
        self.thunder_bag: {} = {}
        self.coordinates: {} = {}
        self.first: str = ''
        self.stats: {} = {}
        self.fit_params: {} = {}
        self.x_ind: Union[None, int] = None
        self.y_ind: Union[None, int] = None
        self.e_ind: Union[None, int] = None
        self.img_path: Union[None, str] = None
        self.map: Union[None, str] = None
        self.datapath: Union[None, str] = None
        self.x_coord_ind: Union[None, int] = None
        self.y_coord_ind: Union[None, int] = None

        if isinstance(input, Thunder):  # if only pass one but its already a thunder object then just use that
            self.thunder_bag[0] = Thunder(input)  # add all the details in depending on args
        elif isinstance(input, dict):
            self.create_bag(input)  # add all the details in depending on args
        else:
            raise TypeError('Cannot convert input to ThunderBag object')

    def create_bag(self, inp):
        self.x_ind =  inp.get('x_ind', 2)
        self.y_ind = inp.get('y_ind', 3)
        self.e_ind = inp.get('e_ind', None)
        self.img_path = inp.get('imgpath', None)
        self.coordinates = inp.get('coords', {})

        self.map = inp.get('map', None) # if user passes map as True then the file will be treated as a map file

        data_paths = inp.get('datapath', None)
        self.datapath = literal_eval(data_paths) # this is a bit dangerous!!!!!

        for i, data in tqdm(enumerate(self.datapath)):
            if len(self.datapath):
                prefix = f'{i}_'  # if more than one datapath then we name them with i_j
            if isinstance(data, Thunder):
                self.thunder_bag[i] = data
            elif isinstance(data, str):
                # then read the data file
                if self.map == True:
                    prefix = ''
                    self.x_coord_ind, self.y_coord_ind = inp.get('x_coord_ind', 0), inp.get('y_coord_ind', 1)
                    map_path = glob(data)[0] # save the filepath to the mapscan as self.map for later
                    x_data, y_data, x_coords, y_coords = self.read_map(map_path, self.x_ind, self.y_ind, self.x_coord_ind, self.y_coord_ind)

                    for j in range(len(x_data)): # go through the list of x_data
                        x_data_, y_data_ = x_data[j], y_data[j] # the x and y data for each coordinate set
                        self.thunder_bag[f'{prefix}{j}'] = Thunder(inp, x_data=x_data_, y_data=y_data_) # make a thunder obj
                                                                                                    # with this data
                        x_coords_, y_coords_ = x_coords[j], y_coords[j]
                        self.coordinates[f'{prefix}{j}'] = (x_coords_, y_coords_)  # for each i we will have a list of tuples of x and y coords
                elif '*' in data:
                    filematches = glob(data)
                    for j, file in enumerate(filematches):
                        try:
                            self.thunder_bag[f'{prefix}{j}'] = self.create_thunder(file, inp) # make a thunder object for each file
                        except ParserError as e:
                            logging.warning(f"A Thunder object could not be created for the datafile: {file}, skipping")
                else:
                    try:
                        self.thunder_bag[str(i)] = self.create_thunder(data, inp)
                    except ParserError as e:
                        logging.warning(f"A Thunder object could not be created for the datafile: {file}, skipping")
            else:
                logging.warning(f"wrong format in data list detected for {i}th element: {data}. Skipping element")
                pass

    @staticmethod
    def create_thunder(file, inp):
        arguments = deepcopy(inp)
        arguments['datapath'] = file
        thund_obj = Thunder(arguments)
        return thund_obj

    @staticmethod
    def read_map(file_address, x_ind, y_ind, x_coord_ind, y_coord_ind):
        x_data, y_data, _ = utili.load_data(file_address, x_ind, y_ind)  # load the data. note these drop nan rows but
                                    # does that for the whole filepath so will be consistent for data and coordinates
        x_coords, y_coords, _ = utili.load_data(file_address, x_coord_ind, y_coord_ind)  # load the coordinates
        x_data, y_data, x_coords, y_coords = utili.map_unique_coords(x_data, y_data, x_coords, y_coords) #

        return x_data, y_data, x_coords, y_coords

    @staticmethod
    def bag_iterator(bag, func, input_args, sett_args):
        bagkeys = tqdm(bag.keys())
        bagkeys.set_description(f"Operating with: {func.__name__}, to find: {sett_args}")
        for key in bagkeys:
            thund = bag.get(key) # bag[key]
            kwargs_ = [getattr(thund, arg) for arg in input_args]
            _, val = utili.apply_func((key, kwargs_), func)
            for i, arg in enumerate(sett_args):
                try:
                    setattr(thund, arg, val[i])
                except KeyError as e:
                    if isinstance(val, dict):
                        setattr(thund, arg, val)
                    else:
                        print(f'Weird KeyError encountered: {e}')

    def choose_spectrum(self):
        # then we have to choose which spectrum we want
        first = next(iter(self.thunder_bag.keys())) # changed from list to iter
        while True:
            try:
                first_thunder = self.thunder_bag[first]
                fig, ax = plt.subplots()
                ax.plot(getattr(first_thunder, 'x_data'), getattr(first_thunder, 'y_data'))
                print(f"Need a decision on which plot is representitive of data, the following is for index {first}")
                plt.show(block=True)
                ans = input("If you are happy with using this data file, type y, otherwise enter a new index")
                if ans == 'y':
                    break
                else:
                    try:
                        first = str(ans)
                    except ValueError:
                        print("You entered an incorrect answer! Trying again...")
            except KeyError:
                print('incorrect key, please enter a lower index value')
        self.first = first

    def clip_data(self):
        first_thunder = self.thunder_bag[self.first]
        clip_left, clip_right = utili.clip_data(getattr(first_thunder, 'x_data'), getattr(first_thunder, 'y_data'))
        for thund in self.thunder_bag.values():
            setattr(thund, 'x_data', getattr(thund, 'x_data')[clip_left:clip_right])
            setattr(thund, 'y_data', getattr(thund, 'y_data')[clip_left:clip_right])

    def bg_param_setter(self):
        # add step to find bg parameters for first one and use for the rest.
        first_thunder = self.thunder_bag[self.first]
        if isinstance(getattr(first_thunder, 'background'), str) and getattr(first_thunder, 'background')=="SCARF":
            _, _, params = bg_remove.background_finder(getattr(first_thunder, 'x_data'), getattr(first_thunder, 'y_data'),
                                                       getattr(first_thunder, 'background'),
                                                       getattr(first_thunder, 'scarf_params'))
            [param.pop('b', None) for param in params]  # we want to find b each time so don't set it for all others
            for thund in self.thunder_bag.values():
                setattr(thund, 'scarf_params', params)  # set all the values to this

    def peak_info_setter(self):
        # add step to find bg parameters for first one and use for the rest.
        first_thunder = self.thunder_bag[self.first]
        no_peaks, peak_centres, peak_amps, peak_widths, peak_types, prominence = \
            peak_finding.find_peak_details(getattr(first_thunder, 'x_data'), getattr(first_thunder, 'y_data_bg_rm'),
                                           getattr(first_thunder, 'no_peaks'),
                                           getattr(first_thunder, 'peak_centres'), getattr(first_thunder, 'peak_amps'),
                                           getattr(first_thunder, 'peak_widths'),
                                           getattr(first_thunder, 'peak_types'))
        for thund in self.thunder_bag.values():  # set these first values for all of them
            setattr(thund, 'no_peaks', no_peaks)  # set values
            setattr(thund, 'peak_centres', peak_centres)  # set values
            setattr(thund, 'peak_types', peak_types)  # set values

    def bound_setter(self, bounds=None):
        if not bounds:
            first_thunder = self.thunder_bag[self.first]
            bounds = peak_finding.make_bounds(getattr(first_thunder, 'tightness'), getattr(first_thunder, 'no_peaks'),
                                              getattr(first_thunder, 'bounds'),
                                              getattr(first_thunder, 'peak_widths'), getattr(first_thunder, 'peak_centres'),
                                              getattr(first_thunder, 'peak_amps'))
        for thund in self.thunder_bag.values():  # set these first values for all of them
            setattr(thund, 'bounds', bounds)  # set values

    def make_fit_params(self):
        fit_params = {}
        first_thunder = self.thunder_bag.get(self.first)
        params = list(getattr(first_thunder, 'peak_params').keys())[
                 : len(
                     getattr(first_thunder, 'peak_params')) // getattr(first_thunder, 'no_peaks')]  # what are the peak params for the first peak
        params = [param.split('_')[1] for param in params]  # keep only the type of param
        for param in params:
            fit_params[param] = {}  # e.g. 'center'
            for key in self.thunder_bag.keys():
                fit_details = getattr(self.thunder_bag.get(key), 'peak_params')
                fit_details = [fit_details.get(key_) for key_ in fit_details.keys() if param in key_]
                fit_params[param][key] = fit_details
        self.fit_params = fit_params

    def get_fit_stats(self):
        stats = {'chisq': {}, 'reduced_chi_sq': {}, 'free_params': {}}
        for key, thund in self.thunder_bag.items():
            chisq = getattr(getattr(thund, 'peaks'), 'chisq')
            reduced_chi_sq = getattr(getattr(thund, 'peaks'), 'redchi')
            free_params = round(chisq / reduced_chi_sq)
            stats['chisq'][key] = chisq
            stats['reduced_chi_sq'][key] = reduced_chi_sq
            stats['free_params'][key] = free_params
        self.stats = stats

    def save_failed_plots(self, dirname):
        for key, thund in self.thunder_bag.items():
            if not getattr(getattr(thund, 'peaks'), 'success'):
                thund.plot_all()
                utili.save_plot(thund.plot, path=dirname,
                                figname=f"failed_plot_{key}_at_position_{self.coordinates.get(key)}.svg")
                thund.plot.close()  # close so memory is conserved.

def main(arguments):
    bag = ThunderBag(deepcopy(arguments)) # load object
    return bag



import logging
from os.path import basename
from ast import literal_eval

from . import utilities as utili
from . import multi_obj
from .background import background_removal as bg_remove
from . import peak_finding
from . import peak_fitting
from . import parsing
from . import map_scan_tools

def main():
    LOGGER = logging.getLogger(__name__)
    LOGGER.setLevel(logging.INFO)

    args = parsing.parse_user_args()

    arguments, LOGGER = parsing.using_user_args(args, LOGGER)

    # fix this. can be moved to then end near dir creation if you sort out the list requirement in multiobj.
    try:  # a user can pass in a list of filenames or just one
        file_name = basename(literal_eval(arguments['datapath'])[0])
    except SyntaxError:  # assume its just a string and not a list passed
        file_name = None
        arguments['datapath'] = f"['{arguments['datapath']}',]"  # as this is what multiobj needs

    bag = multi_obj.main(arguments) # create a Thunder object

    bag.first = list(bag.thunder_bag.keys())[0]
    if arguments.get('clip_data', False) or arguments.get('bg_first_only', False) or arguments.get(
            'peakf_first_only', False) or arguments.get('bounds_first_only', False):
        bag.choose_spectrum() # choose which spectrum to base everything off of if user wants to use one spectra to choose parameters

    ###### clip the data if weird edges
    if arguments.get('clip_data', False):
        bag.clip_data()

    ###### cosmic ray removal goes here
    ####################################################################################################################

    ###### remove background
    if arguments.get('bg_first_only', False):
        bag.bg_param_setter()
    bag.bag_iterator(getattr(bag, 'thunder_bag'), bg_remove.background_finder, ('x_data', 'y_data',
                                                                   'background', 'scarf_params'), ('background', 'y_data_bg_rm', 'params')) # determine the background

    ###### normalisation
    if args.normalise:
        bag.bag_iterator(getattr(bag, 'thunder_bag'), utili.normalise_all, ('y_data_bg_rm', 'background', 'y_data'), ('y_data_bg_rm', 'background', 'y_data_norm'))

    ###### find peaks
    if arguments.get('peakf_first_only', False):
        bag.peak_info_setter()
    bag.bag_iterator(getattr(bag, 'thunder_bag'), peak_finding.find_peak_details, ('x_data', 'y_data_bg_rm', 'no_peaks',
                                                  'peak_centres', 'peak_amps', 'peak_widths',
                                                  'peak_types'), ('no_peaks', 'peak_centres', 'peak_amps', 'peak_widths', 'peak_types', 'prominence')) # find peaks/use them if supplied

    ###### find bounds
    if arguments.get('bounds_first_only', False):
        bag.bound_setter()
    else:
        bounds = {'amps': False, 'centers': False, 'widths': False}
        bag.bound_setter(bounds)
    bag.bag_iterator(getattr(bag, 'thunder_bag'), peak_finding.make_bounds, ('tightness', 'no_peaks', 'bounds', 'peak_widths',
                                              'peak_centres', 'peak_amps'), ('bounds',)) # make bounds

    ###### fit peaks
    bag.bag_iterator(getattr(bag, 'thunder_bag'), peak_fitting.fit_peaks, ('x_data', 'y_data_bg_rm', 'peak_types', 'peak_centres',
                               'peak_amps', 'peak_widths', 'bounds'), ('specs', 'model', 'peak_params', 'peaks')) # fit peaks

    ##### fit params dictionary
    # store all the peak parameters in a dictionary, so the keys are e.g. sigma, center, amplitude, and the values are
    # dictionaries with keys as the run number with values as lists of values for all the peaks for that run
    # this will for now assume the same types of peak for all fits!
    bag.make_fit_params()

    ###### fetch stats etc
    bag.get_fit_stats()

    ###### make directory to save everything in
    file_name, dirname = parsing.make_user_files(arguments, file_name)

    ###### plot map scan
    map_scan_tools.plot_map_scan(bag, getattr(bag, 'fit_params'), dirname)

    # save individual plots for each of the failed fits
    bag.save_failed_plots(dirname)

    ###### put here some code for cluster analysis and pca
    ####################################################################################################################

    # save the bag object and it reports
    utili.save_fit_report(getattr(bag, 'stats'), path=dirname, filename=f"{file_name}_report.json")
    utili.save_fit_report(getattr(bag, 'fit_params'), path=dirname, filename=f"{file_name}_peak_info.json")
    utili.save_thunder(bag, path=dirname, filename=f"{file_name}.d")



from numpy import unique, round, array, nanmin, nanmax
from scipy.sparse import coo_matrix
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('TkAgg')
from mpl_toolkits.axes_grid1 import make_axes_locatable

from . import utilities as utili


##### funcs for plotting
def shift_map_matr(coordinates_array):
    coordinates_array[:, 0] = coordinates_array[:, 0] - min(coordinates_array[:, 0])
    coordinates_array[:, 1] = coordinates_array[:, 1] - min(coordinates_array[:, 1])
    return coordinates_array

def map_scan_plot(coordinates, values):
    no_fits = len(list(values.values())[0])
    figs = []
    axs = []
    for i in range(no_fits):
        X = []
        Y = []
        Z = []
        for key in values.keys():
            x, y = coordinates[key]
            z = values[key][i]
            X.append(x)
            Y.append(y)
            Z.append(z)
        x_step = unique(X)[1] - unique(X)[0]
        xx = (round(X / x_step)).astype(int)
        y_step = unique(Y)[1] - unique(Y)[0]
        yy = round((Y / y_step)).astype(int)
        data_ = coo_matrix((Z, (xx, yy))).toarray()
        f = plt.figure()
        ax = plt.gca()
        magma_cmap = matplotlib.cm.get_cmap('magma')
        magma_cmap.set_bad(color='green')
        im = plt.imshow(data_, cmap=magma_cmap, extent=[min(X), max(X), max(Y), min(Y)], vmin=nanmin(data_),
                        vmax=nanmax(data_))
        plt.xlabel('x coordinates')
        plt.ylabel('y coordinates')
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im, cax=cax)
        figs.append(f)
        axs.append(ax)
    return figs, axs

def plot_map_scan(bag, fit_params, dirname):
    coordinates_array = array(list(getattr(bag, 'coordinates').values()))  # convert coordinates for each point into an array
    coordinates_array = shift_map_matr(
        coordinates_array)  # shift so that each coordinate starts at 0, not normalised
    for i, key in enumerate(getattr(bag, 'coordinates')):
        getattr(bag, 'coordinates')[key] = coordinates_array.tolist()[i]  # reassign in the correct format
    while True:
        plt.close()
        ans = input("making map scans, please input which property you would like to scan. options are:"
                    f"\n {[p_ for p_ in fit_params.keys()]}. or type e to exit")
        if ans == 'e':
            break
        else:
            try:
                p = ans
                plot, ax = map_scan_plot(getattr(bag, 'coordinates'), fit_params.get(p))
                for i, pt in enumerate(plot):
                    try:
                        cents = next(iter(fit_params.get('center').values()))
                        pt.suptitle(f'{p}_{i}_heatmap. peak {i} is centered at: {cents[i]}')
                    except KeyError:
                        print("tried to add label for peak centers onto graph, but couldn't fetch the right variable")
                        pt.suptitle(f'{p}_{i}_heatmap')
            except KeyError:
                p = ''
                ax = None
                plot = None
                print('wrong answer entered, trying again!')
            try:
                for i, pt in enumerate(plot):
                    utili.save_plot(pt, path=dirname, figname=f"{p}_{i}.svg")
            except AttributeError:
                print("Tried to save plot but there is no plot yet! Something wen't wrong in making the plot")
#####


import logging
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)
from numpy import ndarray, sqrt, array, min, float64
from pandas.core.frame import DataFrame
from pandas.core.series import Series
from scipy.signal import savgol_filter
import matplotlib
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt

def rcf(A, rad):
    """
    Popular rolling circle filter (RCF) routine as described by James et al.: https://doi.org/10.1366%2F12-06766
    :param A: np matrix of data
    :param r: radius of circle to filter with
    :return: L, a numpy array of length len(A) which is the locus of the RCF
    """
    if isinstance(A, DataFrame) or isinstance(A, Series):
        A = A.values # convert to numpy array
    elif not isinstance(A, ndarray):
        raise TypeError("Incorrect format passed for A")

    assert isinstance(rad, int) and abs(rad) == rad, "rad needs to be a positive integer"
    assert len(A.shape) == 1, "A needs to be a single column"

    RC = semi_circle_array(rad)
    n = len(A)

    if len(RC) >= n:
        print('something has gone wrong')
        raise RuntimeError("The data has less points than the selected radius!")

    A_sub, RC_sub = generate_sub_matrices(A, rad, RC, n)

    D = generate_diff_matrix(A_sub, RC_sub)

    return D

def semi_circle_array(rad):
    """Generate an array of length 2r+1 with values corresponding to points on a semi circle"""
    RC = array([sqrt(rad**2 - (i - rad)**2) for i in range(2*rad + 1)])
    return RC

def generate_sub_matrices(A, rad, RC, n):
    """
    Generate the sub matrices used in RCF routine
    :param A: Data matrix, length n
    :param rad: radius of RCF filter
    :param RC: np array of length 2r+1 with
    :param n:
    :return: A_sub, RC_sub, lists of same dimensions, sampled according to paper: https://doi.org/10.1366%2F12-06766
    """
    ni = n
    A_sub = []
    RC_sub = []

    for i in range(len(A)): # vectorise this when you can be bothered
        if i < (rad + 1):
            a_sub = A[:(i + rad + 1)]
            rc_sub = RC[(rad - i):(2*rad + 1)]
            A_sub.append(a_sub)
            RC_sub.append(rc_sub)

        elif (rad + 1) <= i <= (n - (rad + 1)):
            a_sub = A[(i - rad):(i + rad + 1)]
            rc_sub = RC
            A_sub.append(a_sub)
            RC_sub.append(rc_sub)

        elif i > (n - (rad + 1)):
            a_sub = A[(i - rad):(n + 1)]
            rc_sub = RC[:(rad + (n - i))]
            A_sub.append(a_sub)
            RC_sub.append(rc_sub)

        else:
            print('something has gone wrong')
            raise RuntimeError("Something has gone wrong and I don't know why")

    return A_sub, RC_sub

def generate_diff_matrix(A_sub, RC_sub):
    """
    Generate a difference matrix, which will be len(A_sub)=len(RC_sub). Each element will be the minimum of matrix
    formed by the difference of AC_sub_i and RC_sub_i at each i (element).
    :param A_sub: List of np arrays
    :param RC_sub: List of np arrays
    :return: D, Difference matrix of shape (len(A_sub),), elements as described above
    """
    D = []
    for i in range(len(A_sub)):
        diff = A_sub[i] - RC_sub[i]
        D.append(min(diff))

    D = array(D)
    return D

def smooth(L, window_length, polyorder):
    """
    Savgol filter applied to data
    :param L: the data to be smoothed
    :param window_length: positive odd integer
    :param polyorder: must be less than window length
    :return:
    """
    return savgol_filter(L, window_length, polyorder, mode='mirror')

def perform_scarf(x_data, y_data, scarf_params=False):

    bg = array([0 for _ in y_data], dtype=float64)
    rad = 70
    window_length, poly_order = 51, 3
    L_sg = 0
    data_bg_rm_y = y_data.copy()
    params = []

    if isinstance(scarf_params, bool) and not scarf_params:
        while True:
            #rcf step
            while True:
                D = rcf(data_bg_rm_y, rad)
                fig, ax = plt.subplots()
                ax.plot(x_data, D)
                ax.plot(x_data, data_bg_rm_y)
                print(f"SCARF background removal requires user input. Please look at the following bg with rad={rad}")
                plt.show(block=True)
                ans = input("If you are happy with the plot, type y. if not then please type a new rad")
                if ans == 'y':
                    break
                else:
                    try:
                        rad = int(ans)
                    except ValueError:
                        print("You entered an incorrect answer! Trying again...")

            L = D
            # used to ask user for bg but now just calculate it after the smoothing step

            # then apply SG filter to L
            while True:
                try:
                    L_sg = smooth(L, window_length, poly_order)
                    fig, ax = plt.subplots()
                    ax.plot(x_data, L_sg)
                    ax.plot(x_data, data_bg_rm_y)
                    print(f"Please look at the following bg with Sg filter parameters (window length, polynomial order): "
                          f"{window_length}, {poly_order}")
                    plt.show(block=True)
                except ValueError as e:
                    print(
                        "Incorrect values for window_length and poly_order have been entered. Poly order must be less than window length and window length must be odd")
                ans = input("please enter y if you are happy with these values, or enter two integers with a space "
                            "for window_length and poly_order")
                if ans == 'y':
                    L = L_sg
                    break
                else:
                    try:
                        ans = ans.split(' ')
                        if len(ans) != 2:
                            raise ValueError("The tuple was more than two elements long")
                        window_length = int(ans[0])
                        poly_order = int(ans[1])
                    except ValueError:
                        print("You entered an incorrect answer! Trying again...")

            # get the bg shift up automatically
            b = min(data_bg_rm_y - L)  # whats the smallest difference between D and b? shift it up by that
            L = L + b

            # final question before exiting
            fig, ax = plt.subplots()
            ax.plot(x_data, L)
            ax.plot(x_data, data_bg_rm_y)
            print(f"Please look at the following bg with selected parameters")
            plt.show(block=True)
            ans = input("Are you happy with this bg? If yes, type y, else type n. n will restart the fitting. \n"
                        "Typing repeat will add an additional bg subtraction to this one ")
            if ans == 'y':
                bg += L
                data_bg_rm_y -= L
                params.append({'rad': rad, 'b': b, 'window_length': window_length, 'poly_order': poly_order})
                break
            elif ans == 'n':
                pass
            elif ans == 'repeat':
                bg += L
                print("apply two bg removal steps, this will mean the background just specified will be removed "
                      "from the data")
                data_bg_rm_y -= L  # remove the bg found here from the original data and go again
                params.append({'rad':rad, 'b':b, 'window_length':window_length, 'poly_order':poly_order})
            else:
                print("You entered an incorrect answer! Trying whole fitting routine again...")
    elif isinstance(scarf_params, bool):
        # the user has passed True! recall this function and change it to false
        logging.warning('scarf has been passed the True bool value for scarf params. This is invalid, assuming you meant '
                        'False as no parameters were specified.')
        data_bg_rm_y, bg, params_ = perform_scarf(x_data, data_bg_rm_y, scarf_params=False)
        params += params_ # params_ is a list of one dictionary
        return data_bg_rm_y, bg
    elif isinstance(scarf_params, dict):
        rad, window_length, poly_order = \
            scarf_params['rad'],  scarf_params['window_length'], scarf_params['poly_order']
        D = rcf(data_bg_rm_y, rad)
        L = smooth(D, window_length, poly_order)
        try:
            b = scarf_params['b'] #  if passed then use it
        except KeyError: # otherwise find it
            b = min(y_data - L) # whats the smallest difference between D and b? shift it up by that
        L = D + b
        bg += L
        data_bg_rm_y -= L
        params.append({'rad': rad, 'b': b, 'window_length': window_length, 'poly_order': poly_order})
    elif isinstance(scarf_params, list):
        # the user wants multiple runs. call the function for each set of params, passing the new y data each time
        for param_dict in scarf_params:
            data_bg_rm_y, bg_, params_ = perform_scarf(x_data, data_bg_rm_y, scarf_params=param_dict)
            bg += bg_
            params += params_ # params_ is a list of a dictionary so use += here
    else:
        logging.warning(
            'an unexpected parameter has been passed as a scarf value, running interactively.')
        data_bg_rm_y, bg, params_ = perform_scarf(x_data, data_bg_rm_y, scarf_params=False)
        params += params_
        return data_bg_rm_y, bg, params
    plt.close()
    return data_bg_rm_y, bg, params


import logging

from . import thundobj
from . import utilities as utili
from .background import background_removal as bg_remove
from . import peak_finding
from . import peak_fitting
from . import parsing


def main():
    LOGGER = logging.getLogger(__name__)
    LOGGER.setLevel(logging.INFO)

    args = parsing.parse_user_args()

    arguments, LOGGER = parsing.using_user_args(args, LOGGER)

    thunder = thundobj.main(arguments) # create a Thunder object

    thunder.background, thunder.y_data_bg_rm, _ = bg_remove.background_finder(thunder.x_data, thunder.y_data,
                                                                           thunder.background, thunder.scarf_params)
                                                                           # determine the background
    if args.normalise:
        thunder.y_data_bg_rm, thunder.background, thunder.y_data_norm = \
                                                 utili.normalise_all(thunder.y_data_bg_rm, thunder.background, thunder.y_data)

    thunder.no_peaks, thunder.peak_centres, thunder.peak_amps, thunder.peak_widths, thunder.peak_types, _ = \
                   peak_finding.find_peak_details(thunder.x_data, thunder.y_data_bg_rm, thunder.no_peaks,
                                                  thunder.peak_centres, thunder.peak_amps, thunder.peak_widths,
                                                  thunder.peak_types) # find peaks/use them if supplied

    thunder.bounds = peak_finding.make_bounds(thunder.tightness, thunder.no_peaks, thunder.bounds, thunder.peak_widths,
                                              thunder.peak_centres, thunder.peak_amps) # make bounds

    thunder.specs, thunder.model, thunder.peak_params, thunder.peaks = \
        peak_fitting.fit_peaks(thunder.x_data, thunder.y_data_bg_rm, thunder.peak_types, thunder.peak_centres,
                               thunder.peak_amps, thunder.peak_widths, thunder.bounds) # fit peaks

    thunder.chi_sq = thunder.peaks.chisqr # set the stats from the fits
    reduced_chi_sq = thunder.peaks.redchi
    thunder.free_params = round(thunder.chi_sq / reduced_chi_sq)

    thunder.plot_all() # plot the data in full and save as an object
    thunder.gen_fit_report() # generate a fit report

    # save a plot of the figure and the thunder object
    file_name, dirname = parsing.make_user_files(arguments, file_name=None)

    utili.save_plot(thunder.plot, path=dirname, figname=f"{file_name}.svg")
    utili.save_thunder(thunder, path=dirname, filename=f"{file_name}.d")
    utili.save_fit_report(thunder.fit_report, path=dirname, filename=f"{file_name}_report.json")